from typing import Dict, List, Literal, Tuple, Type

from pydantic import PrivateAttr

from ....consts import ObjectPositionType
from ....modifiable_values import CostValue
from ....match import Match
from ....card.support.companions import CompanionBase
from ....card.support.locations import LocationBase
from ....card.equipment.artifact.base import ArtifactBase
from ....card.equipment.weapon.base import WeaponBase
from ....action import ActionTypes, Actions
from ....event import UseCardEventArguments
from ....struct import Cost
from .....utils.class_registry import register_class
from .....utils.desc_registry import DescDictType
from ....card.support.items import RoundEffectItemBase


class MementoLens_4_3(RoundEffectItemBase):
    name: Literal["Memento Lens"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number = 1)
    played_card_names: Dict[str, int] = {}
    _accept_card_types: Tuple[Type] = PrivateAttr(
        (WeaponBase, ArtifactBase, LocationBase, CompanionBase)
    )
    max_usage_per_round: int = 1
    # mark USE_CARD as avaliable in deck to record used cards
    available_handler_in_deck: List[ActionTypes] = [
        ActionTypes.USE_CARD, 
    ]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[Actions]:
        """
        If self played an accepted card type, record it.
        """
        if event.action.card_position.id == self.id:
            # self, do normal play
            return super().event_handler_USE_CARD(event, match)
        if event.action.card_position.player_idx != self.position.player_idx:
            # not self play, do nothing
            return []
        card = event.card
        if not isinstance(card, self._accept_card_types):
            # not accepted card type, do nothing
            return []
        self.played_card_names[card.name] = 1
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Match, mode: Literal["REAL", "TEST"]
    ) -> CostValue:
        """
        If self played a card that name in played_card_names, and has usage,
        decrease 2 cost.
        """
        if self.usage <= 0:
            # no usage, do nothing
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            source_area = ObjectPositionType.SUPPORT
        ):
            # not self player or not in support, do nothing
            return value
        card = match.get_object(value.position)
        assert card is not None
        if card.name not in self.played_card_names:
            # not played card, do nothing
            return value
        # decrease cost
        result = [
            value.cost.decrease_cost(value.cost.elemental_dice_color),
            value.cost.decrease_cost(value.cost.elemental_dice_color),
        ]
        if any(result) and mode == 'REAL':
            # decrease success and in real mode
            self.usage -= 1
        return value


desc: Dict[str, DescDictType] = {
    "SUPPORT/Memento Lens": {
        "names": {
            "en-US": "Memento Lens",
            "zh-CN": "留念镜"
        },
        "descs": {
            "4.3": {
                "en-US": "When you play a Weapon, Artifact, Location, or Companion card: If you have played a card with the same name previously during this match, spend 2 less Elemental Dice to play it. (Once per Round)\nUsage(s): 2",  # noqa: E501
                "zh-CN": "我方打出「武器」「圣遗物」「场地」「伙伴」手牌时：如果本场对局中我方曾经打出过所打出牌的同名卡牌，则少花费2个元素骰。（每回合1次）\n可用次数：2"  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Prop_Mirror.png",  # noqa: E501
        "id": 323006
    },
}


register_class(MementoLens_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import CostValue
from ...event import RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)
# Skills


class SearingOnslaught(ElementalSkillBase):
    name: Literal['Searing Onslaught'] = 'Searing Onslaught'
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    counter: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset counter
        """
        self.counter = 0
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['REAL', 'TEST']
    ) -> CostValue:
        """
        If self equipped talent, reduce the second use cost by 1
        """
        if self.counter != 1:
            # not second use
            return value
        if not self.is_talent_equipped(match):
            # no talent
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True, 
            charactor_idx_same = True, area_same = True, id_same = True
        ):
            # not this skill
            return value
        # decrease cost
        value.cost.decrease_cost(DieColor.PYRO)
        return value

    def get_actions(self, match: Any) -> List[Actions]:
        """
        if it is the third time to use, increase damage by 2
        """
        self.counter += 1
        if self.counter == 3:
            self.damage = 5
        ret = super().get_actions(match)
        self.damage = 3
        return ret


class Dawn(ElementalBurstBase):
    name: Literal['Dawn'] = 'Dawn'
    damage: int = 8
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status(
                'Pyro Elemental Infusion',
                { 'mark': 'Diluc' }
            )
        ]


# Talents


class FlowingFlame_3_3(SkillTalent):
    name: Literal['Flowing Flame']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Diluc'] = 'Diluc'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Searing Onslaught'] = 'Searing Onslaught'


# charactor base


class Diluc_3_3(CharactorBase):
    name: Literal['Diluc']  # Do not set default value for charactor name
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | SearingOnslaught | Dawn
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Tempered Sword',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SearingOnslaught(),
            Dawn()
        ]


register_class(Diluc_3_3 | FlowingFlame_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DieColor, ElementType, 
    FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Skills


class InfluxBlast(ElementalSkillBase):
    name: Literal['Influx Blast'] = 'Influx Blast'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        First remove existing Refraction on opposite, then create new
        Refraction.
        """
        target_charactors = match.player_tables[
            1 - self.position.player_idx].charactors
        exist_status = None
        for charactor in target_charactors:
            for status in charactor.status:
                if status.name == 'Refraction':
                    exist_status = status
                    break
            if exist_status is not None:
                break
        ret: List[Actions] = []
        if exist_status:
            ret.append(RemoveObjectAction(
                object_position = exist_status.position
            ))
        args = {}
        if self.is_talent_equipped(match):
            args = {
                'usage': 3,
                'max_usage': 3,
                'is_talent_activated': True,
            }
        return super().get_actions(match) + ret + [
            self.create_opposite_charactor_status(
                match, 'Refraction', args
            )
        ]


# Talents


class MirrorCage_3_3(SkillTalent):
    name: Literal['Mirror Cage']
    version: Literal['3.3']
    charactor_name: Literal['Mirror Maiden'] = 'Mirror Maiden'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4
    )
    skill: Literal['Influx Blast'] = 'Influx Blast'


class MirrorCage_4_2(MirrorCage_3_3):
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
    )


# charactor base


class MirrorMaiden_3_7(CharactorBase):
    name: Literal['Mirror Maiden']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | InfluxBlast | ElementalBurstBase
    ] = []
    faction: List[FactionType] = [
        FactionType.FATUI
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Water Ball',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            InfluxBlast(),
            ElementalBurstBase(
                name = 'Rippled Reflection',
                damage = 5,
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalBurstBase.get_cost(self.element, 3, 2)
            )
        ]


register_class(MirrorMaiden_3_7 | MirrorCage_4_2 | MirrorCage_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...action import (
    Actions, ChangeObjectUsageAction, CreateObjectAction, MakeDamageAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class TenguJuuraiAmbush_3_5(AttackerSummonBase):
    name: Literal['Tengu Juurai: Ambush'] = 'Tengu Juurai: Ambush'
    version: Literal['3.5'] = '3.5'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | CreateObjectAction | ChangeObjectUsageAction]:
        ret: List[
            MakeDamageAction | CreateObjectAction | ChangeObjectUsageAction
        ] = []
        ret += super().event_handler_ROUND_END(event, match)
        active_idx = match.player_tables[
            self.position.player_idx].active_charactor_idx
        ret.append(CreateObjectAction(
            object_name = 'Crowfeather Cover',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                charactor_idx = active_idx,
                area = ObjectPositionType.CHARACTOR_STATUS,
                id = -1
            ),
            object_arguments = {}
        ))
        return ret


class TenguJuuraiStormcluster_3_5(TenguJuuraiAmbush_3_5):
    name: Literal['Tengu Juurai: Stormcluster'] = 'Tengu Juurai: Stormcluster'
    usage: int = 2
    max_usage: int = 2
    damage: int = 2


# Skills


class TenguStormcall(ElementalSkillBase):
    name: Literal['Tengu Stormcall'] = 'Tengu Stormcall'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Tengu Juurai: Ambush')
        ]


class SubjugationKoukouSendou(ElementalBurstBase):
    name: Literal['Subjugation: Koukou Sendou'] = 'Subjugation: Koukou Sendou'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Tengu Juurai: Stormcluster')
        ]


# Talents


class SinOfPride_3_5(SkillTalent):
    name: Literal['Sin of Pride']
    version: Literal['3.5'] = '3.5'
    charactor_name: Literal['Kujou Sara'] = 'Kujou Sara'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: Literal['Subjugation: Koukou Sendou'] = 'Subjugation: Koukou Sendou'


class SinOfPride_4_2(SkillTalent):
    name: Literal['Sin of Pride']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Kujou Sara'] = 'Kujou Sara'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Tengu Stormcall'] = 'Tengu Stormcall'


# charactor base


class KujouSara_3_5(CharactorBase):
    name: Literal['Kujou Sara']
    version: Literal['3.5'] = '3.5'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | TenguStormcall | SubjugationKoukouSendou
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Tengu Bowmanship',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            TenguStormcall(),
            SubjugationKoukouSendou()
        ]


register_class(
    KujouSara_3_5 | SinOfPride_3_5 | SinOfPride_4_2 | TenguJuuraiAmbush_3_5 
    | TenguJuuraiStormcluster_3_5
)



import os
from src.lpsim import MatchState
from tests.utils_for_test import (
    check_hp, check_usage, get_pidx_cidx, get_test_id_from_command, 
    make_respond, set_16_omni, read_from_log_json
)


def test_lyney():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), 'jsons', 'test_lyney.json')
    )
    # modify hp
    for i in range(2):
        charactors = match.player_tables[i].player_deck_information.charactors
        for c in charactors:
            c.hp = c.max_hp = 30
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx, cidx = get_pidx_cidx(cmd)
                status = match.player_tables[pidx].charactors[cidx].status
                check_usage(status, cmd[4:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[4:])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_lyney()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DeclareRoundEndAttackSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, PhysicalNormalAttackBase, 
    CharactorBase, SkillTalent
)


# Summons


class BounatifulCore_4_2(DeclareRoundEndAttackSummonBase):
    name: Literal['Bountiful Core'] = 'Bountiful Core'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 2
    renew_type: Literal['ADD'] = 'ADD'
    extra_attack_usage: int = 2


# Skills


class DanceOfHaftkarsvar(ElementalSkillBase):
    name: Literal['Dance of Haftkarsvar'] = 'Dance of Haftkarsvar'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        all_hydro_dendro = True
        charactors = match.player_tables[self.position.player_idx].charactors
        for c in charactors:
            if (
                c.element != ElementType.HYDRO 
                and c.element != ElementType.DENDRO
            ):
                all_hydro_dendro = False
                break
        ret = super().get_actions(match)
        if all_hydro_dendro:
            # first generate team status, then attack
            ret = [
                self.create_team_status('Golden Chalice\'s Bounty')
            ] + ret
        return ret


class DanceOfAbzendegiDistantDreamsListeningSpring(ElementalBurstBase):
    name: Literal[
        'Dance of Abzendegi: Distant Dreams, Listening Spring'
    ] = 'Dance of Abzendegi: Distant Dreams, Listening Spring'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_opposite_charactor_status(
                match, 'Lingering Aeon', {}
            )
        ]


# Talents


class TheStarrySkiesTheirFlowersRain_4_2(SkillTalent):
    name: Literal['The Starry Skies Their Flowers Rain']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Nilou'] = 'Nilou'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )
    skill: Literal['Dance of Haftkarsvar'] = 'Dance of Haftkarsvar'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        When self Bountiful Core damage, increase damage by 1
        """
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SUMMON,
        ):
            # not equipped, or source not self summon
            return value
        summon = match.get_object(value.position)
        if summon is None or summon.name != 'Bountiful Core':
            # not Bountyful Core
            return value
        # damage +1
        value.damage += 1
        return value


# charactor base


class Nilou_4_2(CharactorBase):
    name: Literal['Nilou']
    version: Literal['4.2'] = '4.2'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | DanceOfHaftkarsvar 
        | DanceOfAbzendegiDistantDreamsListeningSpring
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Dance of Samser',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            DanceOfHaftkarsvar(),
            DanceOfAbzendegiDistantDreamsListeningSpring(),
        ]


register_class(
    Nilou_4_2 | TheStarrySkiesTheirFlowersRain_4_2 | BounatifulCore_4_2
)



from typing import Any, Dict, List, Literal, Type, get_args

from pydantic import PrivateAttr

from ....action import Actions, CreateObjectAction
from .....utils.desc_registry import DescDictType
from .....utils.class_registry import (
    get_class_list_by_base_class, register_class
)
from ....card.support.locations import LocationBase
from ....match import Match
from ....consts import IconType, ObjectPositionType
from ....event import UseCardEventArguments
from ....card.support.items import ItemBase
from ....card.event.foods import FoodCardBase
from ....struct import Cost, ObjectPosition
from ....card.support.base import UsageWithRoundRestrictionSupportBase
from ....card.support.companions import CompanionBase


class Mamere_4_3(CompanionBase, UsageWithRoundRestrictionSupportBase):
    name: Literal['Mamere']
    version: Literal['4.3'] = '4.3'
    cost: Cost = Cost()
    max_usage_one_round: int = 1
    usage: int = 3

    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    _accept_card_types: Type = PrivateAttr(
        LocationBase | FoodCardBase | ItemBase | CompanionBase
    )

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Match
    ) -> List[Actions]:
        # TODO if use food card failed, will mamere generate new card?
        # currently not.
        if not event.use_card_success:
            # use card failed, return
            return []
        if self.position.area != ObjectPositionType.SUPPORT:
            # not placed to support, do original action
            return super().event_handler_USE_CARD(event, match)
        if not self.has_usage():
            # no usage, return
            return []
        if event.action.card_position.player_idx != self.position.player_idx:
            # not self use card, return
            return []
        card = event.card
        if card.name == 'Mamere':
            # using Mamere, return
            return []
        is_target_type = False
        for base_type in get_args(self._accept_card_types):
            if isinstance(card, base_type):
                is_target_type = True
                break
        if is_target_type:
            # using target type, generate new card

            # get candidate list for target type, except self
            kwargs: Dict[str, Any] = { 'exclude': set([self.name]) }
            default_version = match.player_tables[
                self.position.player_idx
            ].player_deck_information.default_version
            if default_version is not None:
                kwargs['version'] = default_version
            candidate_list = get_class_list_by_base_class(
                self._accept_card_types, 
                **kwargs
            )
            random_target = candidate_list[
                int(match._random() * len(candidate_list))
            ]
            # use, and generate new card
            self.use()
            res: List[Actions] = []
            position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.HAND,
                id = -1
            )
            res.append(CreateObjectAction(
                object_name = random_target,
                object_position = position,
                object_arguments = {
                    'version': self.version
                }
            ))
            res += self.check_should_remove()
            return res
        return []


desc: Dict[str, DescDictType] = {
    "SUPPORT/Mamere": {
        "names": {
            "en-US": "Mamere",
            "zh-CN": "玛梅赫"
        },
        "descs": {
            "4.3": {
                "en-US": "After you play a Food/Location/Companion/Item Action Card other than Mamere: Randomly create 1 Food/Location/Companion/Item Action Card other than Mamere, and add it to your hand. (Once per Round).\nUsage(s): 3",  # noqa: E501
                "zh-CN": "我方打出「玛梅赫」以外的「料理」/「场地」/「伙伴」/「道具」行动牌后：随机生成1张「玛梅赫」以外的「料理」/「场地」/「伙伴」/「道具」行动牌，将其加入手牌。（每回合1次）\n可用次数：3"  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_NPC_Mamere.png",  # noqa: E501
        "id": 322021
    },
}


register_class(Mamere_4_3, desc)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost, ObjectPosition

from ...modifiable_values import (
    CostValue, DamageElementEnhanceValue, DamageValue
)

from ...consts import (
    CostLabels, DamageElementalType, DamageType, ElementType, 
    ElementalReactionType, IconType, ObjectPositionType, SkillType
)

from ...action import (
    Actions, ChangeObjectUsageAction, CreateObjectAction, MakeDamageAction, 
    RemoveObjectAction
)

from ...event import (
    AfterMakeDamageEventArguments, ReceiveDamageEventArguments, 
    RoundEndEventArguments, SkillEndEventArguments
)
from .base import (
    CharactorStatusBase, ElementalInfusionCharactorStatus, UsageCharactorStatus
)


class SeedOfSkandha_3_7(UsageCharactorStatus):
    name: Literal['Seed of Skandha'] = 'Seed of Skandha'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        Only seed of charactor received the damage will trigger event, it will
        check all charactor status and trigger them and call ChangeObjectUsage.
        """
        damage_value = event.final_damage
        if damage_value.damage_type != DamageType.DAMAGE:
            # not damage, not trigger
            return []
        if damage_value.element_reaction == ElementalReactionType.NONE:
            # not elemental reaction, not trigger
            return []
        if not self.position.check_position_valid(
            damage_value.target_position, match, 
            player_idx_same = True, charactor_idx_same = True,
        ):
            # damage not received by self, not trigger
            return []
        assert self.usage > 0
        # trigger, check all seed on the same side
        damage_actions: List[MakeDamageAction] = []
        change_usage_actions: List[ChangeObjectUsageAction] = []
        table = match.player_tables[self.position.player_idx]
        has_pyro_charactor = False
        has_talent = False
        # check if enemy has pyro charactor and has talent nahida
        for charactor in match.player_tables[
                1 - self.position.player_idx].charactors:
            if charactor.element == ElementType.PYRO:
                has_pyro_charactor = True
            if charactor.name == 'Nahida':
                if charactor.talent is not None:
                    has_talent = True
        aidx = table.active_charactor_idx
        charactors = table.charactors[aidx:] + table.charactors[:aidx]
        for charactor in charactors:
            for status in charactor.status:
                if status.name == 'Seed of Skandha':
                    # found a seed, trigger it
                    assert status.usage > 0
                    d_ele_type = DamageElementalType.PIERCING
                    if has_pyro_charactor and has_talent:
                        # enemy have talent nahida and pyro, dendeo damage
                        d_ele_type = DamageElementalType.DENDRO
                    # change usage first, so no need to claim new trigger
                    change_usage_actions.append(ChangeObjectUsageAction(
                        object_position = status.position,
                        change_usage = -1,
                    ))
                    damage_actions.append(MakeDamageAction(
                        damage_value_list = [
                            DamageValue(
                                position = status.position,
                                damage = 1,
                                target_position = charactor.position,
                                damage_type = DamageType.DAMAGE,
                                cost = Cost(),
                                damage_elemental_type = d_ele_type,
                            )
                        ],
                    ))
        # firstly change usage, to avoid object is disappeared.
        return change_usage_actions + damage_actions


class VijnanaSuffusion_3_6(ElementalInfusionCharactorStatus, 
                           UsageCharactorStatus):
    name: Literal['Vijnana Suffusion'] = 'Vijnana Suffusion'
    version: Literal['3.6'] = '3.6'
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.DENDRO

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        """
        When character use charged attack, change damage type to dendro
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If talent equipped, and is charged attack, decrease 1 any cost
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.talent is None:
            # no talent, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not this charactor use skill, not modify
            return value
        if not (
            match.player_tables[self.position.player_idx].charge_satisfied
            and value.cost.label & CostLabels.NORMAL_ATTACK.value != 0
        ):
            # not charged attack, not modify
            return value
        # decrease
        value.cost.decrease_cost(None)
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction | RemoveObjectAction]:
        """
        If self use charged attack, summon Clusterbloom Arrow
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
        ):
            # not same player or charactor, or not skill, not modify
            return []
        if not (
            match.player_tables[self.position.player_idx].charge_satisfied
            and match.get_object(event.action.position).skill_type
            == SkillType.NORMAL_ATTACK
        ):
            # not charged attack, not modify
            return []
        assert self.usage > 0
        # summon
        self.usage -= 1
        return [CreateObjectAction(
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.SUMMON,
                id = 0
            ),
            object_name = 'Clusterbloom Arrow',
            object_arguments = {}
        )] + self.check_should_remove()


class RadicalVitality_3_3(CharactorStatusBase):
    name: Literal['Radical Vitality'] = 'Radical Vitality'
    desc: Literal['', 'talent'] = ''
    version: Literal['3.3'] = '3.3'
    usage: int = 0
    max_usage: int = 3
    element_damage_triggered: bool = False
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._update_desc()

    def _update_desc(self) -> None:
        if self.max_usage == 4:
            # talent activated
            self.desc = 'talent'
        else:
            self.desc = ''

    def renew(self, new_status: 'RadicalVitality_3_3') -> None:
        super().renew(new_status)
        self._update_desc()

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == self.max_usage:
            self.usage = 0
        return []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If contains elemental damage and target or source self, gain progress
        """
        satisfied: bool = False
        for damage in event.damages:
            source_self = self.position.check_position_valid(
                damage.final_damage.position, match, 
                player_idx_same = True, charactor_idx_same = True,
            )
            target_self = self.position.check_position_valid(
                damage.final_damage.target_position, match, 
                player_idx_same = True, charactor_idx_same = True,
            )
            if source_self or target_self:
                # source or target self
                if damage.final_damage.damage_type != DamageType.DAMAGE:
                    # heal, do nothing
                    continue
                if (
                    damage.final_damage.damage_elemental_type in [
                        DamageElementalType.PHYSICAL, 
                        DamageElementalType.PIERCING,
                        DamageElementalType.HEAL
                    ]
                ):
                    # physical or piercing or heal, do nothing
                    continue
                if source_self and (
                    damage.final_damage.cost.label 
                    & CostLabels.ELEMENTAL_BURST.value != 0
                ):
                    # source self, and elemental burst, do nothing
                    continue
                satisfied = True
        if satisfied:
            # increase usage
            self.usage = min(self.usage + 1, self.max_usage)
        return []


register_class(SeedOfSkandha_3_7 | VijnanaSuffusion_3_6 | RadicalVitality_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackAndGenerateStatusSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import Actions, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class FierySanctumField_4_1(AttackAndGenerateStatusSummonBase):
    name: Literal['Fiery Sanctum Field'] = 'Fiery Sanctum Field'
    version: Literal['4.1'] = '4.1'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1


# Skills


class MoltenInferno(ElementalSkillBase):
    name: Literal['Molten Inferno'] = 'Molten Inferno'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_name = 'Fiery Sanctum Field'
        summon_exists = False
        for summon in summons:
            if summon.name == summon_name:
                summon_exists = True
                break
        if summon_exists:
            # make damage and create summon
            return super().get_actions(match) + [
                self.create_summon(summon_name),
            ]
        return [
            self.create_summon(summon_name),
            self.charge_self(1)
        ]


class LeonineBite(ElementalBurstBase):
    name: Literal['Leonine Bite'] = 'Leonine Bite'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status('Incineration Drive'),
        ]


class IncinerationDrive(ElementalBurstBase):
    name: Literal['Incineration Drive'] = 'Incineration Drive'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        Always invalid for prepare skills
        """
        return False


# Talents


class StalwartAndTrue_4_1(SkillTalent):
    name: Literal['Stalwart and True']
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Dehya'] = 'Dehya'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
    )
    skill: Literal['Molten Inferno'] = 'Molten Inferno'

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If equipped, and dehya hp <= 6, heal 2 hp
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.hp <= 6:
            return [MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.HEAL,
                        target_position = charactor.position,
                        damage = -2,
                        damage_elemental_type = DamageElementalType.HEAL,
                        cost = self.cost.copy()
                    )
                ],
            )]
        return []


# charactor base


class Dehya_4_1(CharactorBase):
    name: Literal['Dehya']
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | MoltenInferno | LeonineBite 
        | IncinerationDrive
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU, FactionType.THE_EREMITES
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Sandstorm Assault',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            MoltenInferno(),
            LeonineBite(),
            IncinerationDrive(),
        ]


register_class(Dehya_4_1 | StalwartAndTrue_4_1 | FierySanctumField_4_1)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....consts import ELEMENT_TO_DIE_COLOR, ObjectPositionType

from ....action import CreateDiceAction, CreateObjectAction

from ....event import ReceiveDamageEventArguments, RoundPrepareEventArguments

from ....struct import Cost
from .base import ArtifactBase, RoundEffectArtifactBase


class GeneralsAncientHelm_3_5(ArtifactBase):
    name: Literal["General's Ancient Helm"]
    version: Literal['3.5'] = '3.5'
    cost: Cost = Cost(same_dice_number = 2)
    usage: int = 0

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        Create Unmovable Mountain for this charactor.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        position = self.position.set_area(ObjectPositionType.CHARACTOR_STATUS)
        return [CreateObjectAction(
            object_name = 'Unmovable Mountain',
            object_position = position,
            object_arguments = {}
        )]


class TenacityOfTheMillelith_3_7(RoundEffectArtifactBase):
    name: Literal['Tenacity of the Millelith']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)
    max_usage_per_round: int = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        super().event_handler_ROUND_PREPARE(event, match)
        position = self.position.set_area(ObjectPositionType.CHARACTOR_STATUS)
        return [CreateObjectAction(
            object_name = 'Unmovable Mountain',
            object_position = position,
            object_arguments = {}
        )]

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When this charactor received damage and is active charactor, create
        elemental die.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        if self.usage == 0:
            # no usage
            return []
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.hp == 0:
            # charactor is dying
            return []
        damage = event.final_damage
        if not self.position.check_position_valid(
            damage.target_position, match, player_idx_same = True,
            charactor_idx_same = True, source_is_active_charactor = True
        ):
            # damage not attack self, or self not active charactor
            return []
        # create die
        self.usage -= 1
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = ELEMENT_TO_DIE_COLOR[charactor.element]
        )]


register_class(GeneralsAncientHelm_3_5 | TenacityOfTheMillelith_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue, DamageIncreaseValue
from ...event import RoundPrepareEventArguments

from ...action import Actions
from ...struct import Cost

from ...consts import (
    CostLabels, DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class SolarIsotoma_4_0(AttackerSummonBase):
    name: Literal['Solar Isotoma'] = 'Solar Isotoma'
    version: Literal['4.0'] = '4.0'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1

    attack_cost_decrease_usage: int = 1

    def renew(self, match: Any) -> None:
        super().renew(match)
        self.attack_cost_decrease_usage = 1

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round prepare, reset usage
        """
        self.attack_cost_decrease_usage = 1
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        if self use plunging attack and has usage, decrease cost
        """
        if (
            self.attack_cost_decrease_usage > 0
            and value.cost.label & CostLabels.PLUNGING_ATTACK > 0
            and value.position.player_idx == self.position.player_idx
        ):
            # is self and has usage and plunging attack
            if value.cost.decrease_cost(None):
                if mode == 'REAL':
                    self.attack_cost_decrease_usage -= 1
        return value

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['REAL', 'TEST']
    ) -> DamageIncreaseValue:
        """
        If has albedo equip talent, increase plunging attack by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not self
            return value
        found_talent_albedo = False
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.name == 'Albedo' and charactor.talent is not None:
                found_talent_albedo = True
                break
        if not found_talent_albedo:
            # not found
            return value
        if (
            value.cost.label & CostLabels.NORMAL_ATTACK.value > 0
            and match.player_tables[self.position.player_idx].plunge_satisfied
        ):
            # plunging attack
            assert mode == 'REAL'
            value.damage += 1
        return value


# Skills


class AbiogenesisSolarIsotoma(ElementalSkillBase):
    name: Literal['Abiogenesis: Solar Isotoma'] = 'Abiogenesis: Solar Isotoma'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.create_summon('Solar Isotoma'),
            self.charge_self(1),
        ]


class RiteOfProgenitureTectonicTide(ElementalBurstBase):
    name: Literal[
        'Rite of Progeniture: Tectonic Tide'
    ] = 'Rite of Progeniture: Tectonic Tide'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if Solar Isotoma exists. If so, change self.damage to 8.
        """
        summons = match.player_tables[self.position.player_idx].summons
        isotoma_exist: bool = False
        for s in summons:
            if s.name == 'Solar Isotoma':
                isotoma_exist = True
                break
        if isotoma_exist:
            # temporary change damage to 6
            self.damage = 6
        ret = super().get_actions(match)
        self.damage = 4
        return ret


# Talents


class DescentOfDivinity_4_0(SkillTalent):
    name: Literal['Descent of Divinity']
    version: Literal['4.0'] = '4.0'
    charactor_name: Literal['Albedo'] = 'Albedo'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3
    )
    skill: Literal['Abiogenesis: Solar Isotoma'] = 'Abiogenesis: Solar Isotoma'


# charactor base


class Albedo_4_0(CharactorBase):
    name: Literal['Albedo']
    version: Literal['4.0'] = '4.0'
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | AbiogenesisSolarIsotoma 
        | RiteOfProgenitureTectonicTide
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Favonius Bladework - Weiss',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            AbiogenesisSolarIsotoma(),
            RiteOfProgenitureTectonicTide()
        ]


register_class(Albedo_4_0 | DescentOfDivinity_4_0 | SolarIsotoma_4_0)



from typing import Literal

from .....utils.class_registry import register_class

from ....consts import WeaponType
from .base import WeaponBase
from ....struct import Cost


class VanillaWeapon_3_3(WeaponBase):
    name: Literal[
        'Magic Guide',
        'Raven Bow',
        "Traveler's Handy Sword",
        'White Iron Greatsword',
        'White Tassel'
    ]
    version: Literal['3.3'] = '3.3'
    weapon_type: WeaponType = WeaponType.OTHER

    cost: Cost = Cost(same_dice_number = 2)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == 'Magic Guide':
            self.weapon_type = WeaponType.CATALYST
        elif self.name == 'Raven Bow':
            self.weapon_type = WeaponType.BOW
        elif self.name == "Traveler's Handy Sword":
            self.weapon_type = WeaponType.SWORD
        elif self.name == 'White Iron Greatsword':
            self.weapon_type = WeaponType.CLAYMORE
        else:
            assert self.name == 'White Tassel'
            self.weapon_type = WeaponType.POLEARM


register_class(VanillaWeapon_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Skills


class FireworkFlareUp(PhysicalNormalAttackBase):
    name: Literal['Firework Flare-Up'] = 'Firework Flare-Up'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.PYRO)


class NiwabiFireDance(ElementalSkillBase):
    name: Literal['Niwabi Fire-Dance'] = 'Niwabi Fire-Dance'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 1
    )

    def get_actions(self, match: Any) -> List[CreateObjectAction]:
        """
        only create object
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        talent = charactor.talent
        max_usage = 2
        if talent is not None:
            max_usage = talent.status_max_usage
        return [self.create_charactor_status('Niwabi Enshou', {
            'usage': max_usage,
            'max_usage': max_usage
        })]


class RyuukinSaxifrage(ElementalBurstBase):
    name: Literal['Ryuukin Saxifrage'] = 'Ryuukin Saxifrage'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status('Aurous Blaze')
        ]


# Talents


class NaganoharaMeteorSwarm_4_2(SkillTalent):
    name: Literal['Naganohara Meteor Swarm']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Yoimiya'] = 'Yoimiya'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 2,
    )
    skill: Literal['Niwabi Fire-Dance'] = 'Niwabi Fire-Dance'
    status_max_usage: int = 3


# charactor base


class Yoimiya_3_8(CharactorBase):
    name: Literal['Yoimiya']  # Do not set default value for charactor name
    version: Literal['3.8'] = '3.8'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            FireworkFlareUp(),
            NiwabiFireDance(),
            RyuukinSaxifrage()
        ]


register_class(Yoimiya_3_8 | NaganoharaMeteorSwarm_4_2)



from typing import Dict, List, Literal

from ....status.team_status.base import RoundTeamStatus
from .....utils.class_registry import register_class
from ....action import (
    Actions, ChangeObjectUsageAction, CreateObjectAction, DrawCardAction, 
    MakeDamageAction
)
from ....event import (
    RoundEndEventArguments, RoundPrepareEventArguments, SkillEndEventArguments
)
from ....summon.base import AttackerSummonBase
from ....modifiable_values import DamageIncreaseValue
from ....match import Match
from ....struct import Cost, ObjectPosition
from ....consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    IconType, ObjectPositionType, WeaponType
)
from ....charactor.charactor_base import (
    CharactorBase, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, SkillTalent
)
from .....utils.desc_registry import DescDictType


class GeneralsWarBanner_4_3(RoundTeamStatus):
    name: Literal["General's War Banner"] = "General's War Banner"
    version: Literal['4.3'] = '4.3'
    usage: int = 2
    max_usage: int = 3

    icon_type: IconType = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if (
            value.position.area != ObjectPositionType.SKILL
            and value.position.area != ObjectPositionType.CHARACTOR
        ):
            # not damage from charactor
            return value
        if value.position.player_idx != self.position.player_idx:
            # not damage from self
            return value
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not geo damage
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


class GeneralsGlory_4_3(AttackerSummonBase):
    name: Literal["General's Glory"] = "General's Glory"
    version: Literal['4.3'] = '4.3'
    usage: int = 2
    max_usage: int = 2
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction | CreateObjectAction]:
        ret: List[
            MakeDamageAction | ChangeObjectUsageAction | CreateObjectAction
        ] = []
        ret += super().event_handler_ROUND_END(event, match)
        charactors = match.player_tables[self.position.player_idx].charactors
        geo_number = 0
        for charactor in charactors:
            if charactor.element == ElementType.GEO:
                geo_number += 1
        if geo_number >= 2:
            ret.append(CreateObjectAction(
                object_name = 'Crystallize',
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1
                ),
                object_arguments = {}
            ))
        return ret


class InuzakaAllRoundDefense(ElementalSkillBase):
    name: Literal["Inuzaka All-Round Defense"] = "Inuzaka All-Round Defense"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('General\'s War Banner')
        ]


class JuugaForwardUntoVictory(ElementalBurstBase):
    name: Literal[
        "Juuga: Forward Unto Victory"] = "Juuga: Forward Unto Victory"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('General\'s War Banner'),
            self.create_summon('General\'s Glory')
        ]


class RushingHoundSwiftAsTheWind_4_3(SkillTalent):
    name: Literal[
        "Rushing Hound: Swift as the Wind"
    ] = "Rushing Hound: Swift as the Wind"
    version: Literal['4.3'] = '4.3'
    charactor_name: Literal["Gorou"] = "Gorou"
    skill: Literal["Inuzaka All-Round Defense"] = "Inuzaka All-Round Defense"
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
    )
    usage: int = 1
    draw_card: bool = False

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        Reset draw_card_usage
        """
        self.usage = 1
        self.draw_card = False
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Match,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return value
        if (
            value.position.area != ObjectPositionType.SKILL
            and value.position.area != ObjectPositionType.CHARACTOR
        ):
            # not damage from charactor
            return value
        if value.position.player_idx != self.position.player_idx:
            # not damage from self
            return value
        if (
            value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not geo damage
            return value
        if self.usage <= 0:
            # no usage
            return value
        assert mode == 'REAL'
        self.draw_card = True
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[DrawCardAction]:
        """
        If draw_card is True, and has General's War Banner in play,
        and has usage, draw a card and reset draw_card to False.
        """
        if not self.draw_card:
            return []
        self.draw_card = False
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if status.name == 'General\'s War Banner':
                break
        else:
            return []
        assert self.usage > 0
        self.usage -= 1
        return [DrawCardAction(
            player_idx = self.position.player_idx,
            number = 1,
            draw_if_filtered_not_enough = True
        )]


class Gorou_4_3(CharactorBase):
    name: Literal["Gorou"]
    version: Literal['4.3']
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | InuzakaAllRoundDefense 
        | JuugaForwardUntoVictory
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Ripping Fang Fletching',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            InuzakaAllRoundDefense(),
            JuugaForwardUntoVictory(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTOR/Gorou": {
        "names": {
            "en-US": "Gorou",
            "zh-CN": "五郎"
        },
        "image_path": "cardface/Char_Avatar_Gorou.png",  # noqa: E501
        "id": 1606,
        "descs": {
            "4.3": {
                "en-US": "",
                "zh-CN": ""
            }
        }
    },
    "SKILL_Gorou_NORMAL_ATTACK/Ripping Fang Fletching": {
        "names": {
            "en-US": "Ripping Fang Fletching",
            "zh-CN": "呲牙裂扇箭"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Physical DMG.",
                "zh-CN": "造成2点物理伤害。"
            }
        }
    },
    "SKILL_Gorou_ELEMENTAL_SKILL/Inuzaka All-Round Defense": {
        "names": {
            "en-US": "Inuzaka All-Round Defense",
            "zh-CN": "犬坂吠吠方圆阵"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Geo DMG, creates 1 General's War Banner.",
                "zh-CN": "造成2点岩元素伤害，生成大将旗指物。"
            }
        }
    },
    "TEAM_STATUS/General's War Banner": {
        "names": {
            "en-US": "General's War Banner",
            "zh-CN": "大将旗指物"
        },
        "descs": {
            "4.3": {
                "en-US": "Your party's characters deal +1 Geo DMG.\nDuration (Rounds): 2 (Can stack, Can stack till 3 Rounds)",  # noqa: E501
                "zh-CN": "我方角色造成的岩元素伤害+1。\n持续回合：2（可叠加，最多叠加到3回合）"  # noqa: E501
            }
        },
        "image_path": "status/Gorou_E.png"
    },
    "SKILL_Gorou_ELEMENTAL_BURST/Juuga: Forward Unto Victory": {
        "names": {
            "en-US": "Juuga: Forward Unto Victory",
            "zh-CN": "兽牙逐突形胜战法"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Geo DMG, creates 1 General's War Banner, summons General's Glory.",  # noqa: E501
                "zh-CN": "造成2点岩元素伤害，生成大将旗指物，召唤大将威仪。"
            }
        }
    },
    "SUMMON/General's Glory": {
        "names": {
            "en-US": "General's Glory",
            "zh-CN": "大将威仪"
        },
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Geo DMG. If your party has 2 Geo characters, create 1 Crystallize.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：造成1点岩元素伤害；如果队伍中存在2名岩元素角色，则生成结晶。\n可用次数：2"  # noqa: E501
            }
        }
    },
    "TALENT_Gorou/Rushing Hound: Swift as the Wind": {
        "names": {
            "en-US": "Rushing Hound: Swift as the Wind",
            "zh-CN": "犬奔·疾如风"
        },
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Gorou, equip this card.\nAfter Gorou equips this card, immediately use Inuzaka All-Round Defense once.\nWhen your Gorou, who has this card equipped, is on the field, and after any of your characters deals Geo DMG: If General's War Banner is in play, draw 1 card. (Once per Round)\n(You must have Gorou in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为五郎时，装备此牌。\n五郎装备此牌后，立刻使用一次犬坂吠吠方圆阵。\n装备有此牌的五郎在场时，我方角色造成岩元素伤害后：如果场上存在大将旗指物，抓1张牌。（每回合1次）\n（牌组中包含五郎，才能加入牌组）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Gorou.png",  # noqa: E501
        "id": 216061
    },
}


register_class(
    GeneralsGlory_4_3 | GeneralsWarBanner_4_3 | RushingHoundSwiftAsTheWind_4_3 
    | Gorou_4_3, desc
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class TalismanSpirit_3_7(AttackerSummonBase):
    name: Literal['Talisman Spirit'] = 'Talisman Spirit'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        When enemy receives physical or cryo damage, regardless of source,
        increase damage by 1
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.target_position.player_idx == self.position.player_idx:
            # attack self, no nothing
            return value
        if value.damage_elemental_type not in [
            DamageElementalType.CRYO, DamageElementalType.PHYSICAL
        ]:
            # not cryo or physical, do nothing
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


# Skills


class SpringSpiritSummoning(ElementalSkillBase):
    name: Literal['Spring Spirit Summoning'] = 'Spring Spirit Summoning'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )
    max_usage: int = 3

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {
            'usage': self.max_usage,
            'max_usage': self.max_usage,
        }
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if charactor.talent is not None:
            args.update({
                'talent_usage': 1,
                'talent_max_usage': 1,
            })
        return super().get_actions(match) + [
            self.create_team_status('Icy Quill', args),
        ]


class DivineMaidensDeliverance(ElementalBurstBase):
    name: Literal[
        "Divine Maiden's Deliverance"] = "Divine Maiden's Deliverance"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Talisman Spirit')
        ]


# Talents


class MysticalAbandon_3_7(SkillTalent):
    name: Literal['Mystical Abandon'] = 'Mystical Abandon'
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Shenhe'] = 'Shenhe'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )
    skill: Literal['Spring Spirit Summoning'] = 'Spring Spirit Summoning'


# charactor base


class Shenhe_3_7(CharactorBase):
    name: Literal['Shenhe']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase
        | SpringSpiritSummoning | DivineMaidensDeliverance
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Dawnstar Piercer',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SpringSpiritSummoning(),
            DivineMaidensDeliverance(),
        ]


register_class(Shenhe_3_7 | MysticalAbandon_3_7 | TalismanSpirit_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_base_class, register_class

from ...action import (
    ActionTypes, Actions, CharactorReviveAction, CreateObjectAction, 
    MakeDamageAction
)
from ...struct import Cost, ObjectPosition
from ...modifiable_values import DamageValue
from ...consts import (
    CostLabels, DamageElementalType, DamageType, ObjectPositionType
)
from ...object_base import EventCardBase


class FoodCardBase(EventCardBase):
    """
    Base class for food cards.
    """
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.FOOD.value | CostLabels.EVENT.value
    )
    can_eat_only_if_damaged: bool

    def eat_target(self, match: Any) -> List[int]:
        """
        Get target charactor idxs that can eat.
        """
        ret: List[int] = []
        table = match.player_tables[self.position.player_idx]
        for idx, charactor in enumerate(table.charactors):
            if charactor.is_defeated:
                continue
            is_full = False
            for status in charactor.status:
                if status.name == 'Satiated':
                    is_full = True
                    break
            if is_full:
                continue
            if self.can_eat_only_if_damaged and charactor.damage_taken <= 0:
                continue
            ret.append(idx)
        return ret

    def is_valid(self, match: Any) -> bool:
        """
        If there is at least one target that can eat.
        """
        return len(self.eat_target(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Get targets of this card.
        """
        ret: List[ObjectPosition] = []
        charactors = match.player_tables[self.position.player_idx].charactors
        for idx in self.eat_target(match):
            ret.append(charactors[idx].position)
        return ret

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Add Satiated status to target charactor.
        """
        assert target is not None
        pos = target.set_area(ObjectPositionType.CHARACTOR_STATUS)
        return [CreateObjectAction(
            object_name = 'Satiated',
            object_position = pos,
            object_arguments = {}
        )]


register_base_class(FoodCardBase)


class AllCharactorFoodCard(FoodCardBase):
    """
    Food card that can be eaten by all charactors.
    """

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        As it fulfills all charactors, no need to specify target.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[Actions]:
        """
        It will add status to all charactors that is not satiated.
        """
        ret: List[Actions] = []
        targets = self.eat_target(match)
        assert len(targets) > 0
        charactors = match.player_tables[self.position.player_idx].charactors
        for t in targets:
            charactor = charactors[t]
            ret += self.one_charactor_action(charactor.position)
        return ret

    def one_charactor_action(self, pos: ObjectPosition) -> List[Actions]:
        """
        Actions that will be applied to one charactor. Default is to add 
        Satiated and status that has same name as the card.
        """
        ret: List[Actions] = []
        pos = pos.set_area(
            ObjectPositionType.CHARACTOR_STATUS)
        ret.append(CreateObjectAction(
            object_name = 'Satiated',
            object_position = pos,
            object_arguments = {}
        ))
        ret.append(CreateObjectAction(
            object_name = self.name,
            object_position = pos,
            object_arguments = {}
        ))
        return ret


class JueyunGuoba_3_3(FoodCardBase):
    name: Literal['Jueyun Guoba']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        ret.append(action_2)
        return ret


class AdeptusTemptation_3_3(FoodCardBase):
    name: Literal["Adeptus' Temptation"]
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(any_dice_number = 2)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        ret.append(action_2)
        return ret


class LotusFlowerCrisp_3_3(FoodCardBase):
    name: Literal['Lotus Flower Crisp']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        ret.append(action_2)
        return ret


class NorthernSmokedChicken_3_3(FoodCardBase):
    name: Literal['Northern Smoked Chicken']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        ret.append(action_2)
        return ret


class SweetMadame_3_3(FoodCardBase):
    name: Literal['Sweet Madame']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = target,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        ))
        return ret


class MondstadtHashBrown_3_3(FoodCardBase):
    name: Literal['Mondstadt Hash Brown']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = target,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        ))
        return ret


class MushroomPizza_3_3(FoodCardBase):
    name: Literal['Mushroom Pizza']
    version: Literal['3.3'] = '3.3'
    cost: Cost = Cost(same_dice_number = 1)

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = target,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        ))
        create_1 = ret[0]
        assert create_1.type == ActionTypes.CREATE_OBJECT
        ret.append(CreateObjectAction(
            object_name = self.name,
            object_position = create_1.object_position,
            object_arguments = {}
        ))
        return ret


class MintyMeatRolls_3_4(FoodCardBase):
    name: Literal['Minty Meat Rolls']
    version: Literal['3.4'] = '3.4'
    cost: Cost = Cost(same_dice_number = 1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        action_2.object_arguments['version'] = self.version
        ret.append(action_2)
        return ret


class MintyMeatRolls_3_3(MintyMeatRolls_3_4):
    version: Literal['3.3']


class TeyvatFriedEgg_4_1(FoodCardBase):
    name: Literal['Teyvat Fried Egg']
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 2)

    can_eat_only_if_damaged: bool = False

    def is_valid(self, match: Any) -> bool:
        """
        Has target, and not used this round.
        """
        for s in match.player_tables[self.position.player_idx].team_status:
            if s.name == 'Revive on cooldown':
                return False
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Can only target defeated charactors
        """
        return [
            charactor.position for charactor 
            in match.player_tables[self.position.player_idx].charactors 
            if charactor.is_defeated
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | CharactorReviveAction]:
        ret: List[CreateObjectAction | CharactorReviveAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        create_1 = ret[0]
        assert create_1.type == ActionTypes.CREATE_OBJECT
        ret.append(CreateObjectAction(
            object_name = 'Revive on cooldown',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = 0,
            ),
            object_arguments = {}
        ))
        ret = [
            CharactorReviveAction(
                player_idx = target.player_idx,
                charactor_idx = target.charactor_idx,
                revive_hp = 1
            )
        ] + ret
        return ret


class TeyvatFriedEgg_3_7(TeyvatFriedEgg_4_1):
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)


class SashimiPlatter_3_7(FoodCardBase):
    name: Literal['Sashimi Platter']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        action_2 = ret[0].copy(deep = True)
        action_2.object_name = self.name
        ret.append(action_2)
        return ret


class TandooriRoastChicken_3_7(AllCharactorFoodCard):
    name: Literal['Tandoori Roast Chicken']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(any_dice_number = 2)

    can_eat_only_if_damaged: bool = False


class ButterCrab_3_7(AllCharactorFoodCard):
    name: Literal['Butter Crab']
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(any_dice_number = 2)

    can_eat_only_if_damaged: bool = False


register_class(
    JueyunGuoba_3_3 | AdeptusTemptation_3_3 | LotusFlowerCrisp_3_3 
    | NorthernSmokedChicken_3_3 | SweetMadame_3_3 | MondstadtHashBrown_3_3 
    | MushroomPizza_3_3 | MintyMeatRolls_3_4 | MintyMeatRolls_3_3
    | TeyvatFriedEgg_4_1 | TeyvatFriedEgg_3_7 | SashimiPlatter_3_7 
    | TandooriRoastChicken_3_7 | ButterCrab_3_7
)



from typing import Any, Literal

from .....utils.class_registry import register_class

from ....consts import ObjectPositionType, SkillType, WeaponType
from ....modifiable_values import DamageIncreaseValue
from ....struct import Cost
from .base import RoundEffectWeaponBase


class SkywardBase(RoundEffectWeaponBase):
    name: str
    cost: Cost = Cost(same_dice_number = 3)
    version: str
    weapon_type: WeaponType
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return value
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        if value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ) and self.usage > 0:
            # have usage and is normal attack
            assert mode == 'REAL'
            self.usage -= 1
            value.damage += 1
        return value


class SkywardAtlas_3_3(SkywardBase):
    name: Literal['Skyward Atlas']
    version: Literal['3.3'] = '3.3'
    weapon_type: Literal[WeaponType.CATALYST] = WeaponType.CATALYST


class SkywardHarp_3_3(SkywardBase):
    name: Literal['Skyward Harp']
    version: Literal['3.3'] = '3.3'
    weapon_type: Literal[WeaponType.BOW] = WeaponType.BOW


class SkywardSpine_3_3(SkywardBase):
    name: Literal['Skyward Spine']
    version: Literal['3.3'] = '3.3'
    weapon_type: Literal[WeaponType.POLEARM] = WeaponType.POLEARM


class SkywardBlade_3_7(SkywardBase):
    name: Literal['Skyward Blade']
    version: Literal['3.7'] = '3.7'
    weapon_type: Literal[WeaponType.SWORD] = WeaponType.SWORD


class SkywardPride_3_7(SkywardBase):
    name: Literal['Skyward Pride']
    version: Literal['3.7'] = '3.7'
    weapon_type: Literal[WeaponType.CLAYMORE] = WeaponType.CLAYMORE


register_class(
    SkywardAtlas_3_3 | SkywardHarp_3_3 | SkywardPride_3_7 | SkywardSpine_3_3 
    | SkywardBlade_3_7
)



from src.lpsim.server.deck import Deck
from tests.utils_for_test import remove_ids


def test_deck_string():
    deck_dict_1 = Deck(**{
        'name': 'Deck',
        'charactors': [
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'ElectroMobMage',
                'element': 'ELECTRO',
            },
        ],
        'cards': [
            {
                'name': 'Strategize',
            }
        ] * 30,
        'default_version': '4.0',
    })
    deck_str_1 = Deck.from_str('''
        default_version:4.0
        charactor:DendroMobMage*2
        charactor:ElectroMobMage
        Strategize*30

    ''')
    deck_dict_1 = remove_ids(deck_dict_1)
    deck_str_1 = remove_ids(deck_str_1)
    assert deck_str_1 == deck_dict_1
    deck_dict_2 = Deck(**{
        'name': 'Deck',
        'charactors': [
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'ElectroMobMage',
                'element': 'ELECTRO',
            },
        ],
        'cards': [
            {
                'name': 'Wine-Stained Tricorne',
            }
        ] * 10 + [
            {
                'name': 'Laurel Coronet',
            }
        ] * 10 + [
            {
                'name': 'Strategize',
            }
        ] * 10,
        'default_version': '4.0',
    })
    deck_str_2 = Deck.from_str('''
        default_version:4.0

        # charactors
        charactor:DendroMobMage*2
        charactor:ElectroMobMage

        # cards
        Wine-Stained Tricorne*10
        Laurel Coronet*10
        Strategize*10
    ''')
    deck_dict_2 = remove_ids(deck_dict_2)
    deck_str_2 = remove_ids(deck_str_2)
    assert deck_str_2 == deck_dict_2
    deck_dict_3 = Deck(**{
        'name': 'Deck',
        'charactors': [
            {
                'name': 'Fischl',
            },
            {
                'name': 'DendroMobMage',
                'element': 'DENDRO',
            },
            {
                'name': 'PyroMobMage',
                'element': 'PYRO',
            },
        ],
        'cards': [
            {
                'name': 'Stellar Predator',
            }
        ] * 30,
        'default_version': '4.0',
    })
    deck_str_3 = Deck.from_str('''
        default_version:4.0
        charactor:Fischl
        charactor:DendroMobMage
        charactor:PyroMobMage
        Stellar Predator*30
    ''')
    deck_dict_3 = remove_ids(deck_dict_3)
    deck_str_3 = remove_ids(deck_str_3)
    assert deck_str_3 == deck_dict_3


def test_deck_assertions():
    deck = Deck.from_str('''
        default_version:4.0
        charactor:DendroMobMage*2
        charactor:ElectroMobMage
        Strategize*30
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False
    )[0]
    assert deck.check_legal(
        card_number = 30, 
        max_same_card_number = 30,
        charactor_number = 3,
        check_restriction = False
    )[0]
    assert not deck.check_legal(
        card_number = 29, 
        max_same_card_number = 30,
        charactor_number = 3,
        check_restriction = False
    )[0]
    assert not deck.check_legal(
        card_number = 31, 
        max_same_card_number = 30,
        charactor_number = 3,
        check_restriction = False
    )[0]
    assert not deck.check_legal(
        card_number = 30, 
        max_same_card_number = 29,
        charactor_number = 3,
        check_restriction = False
    )[0]
    assert deck.check_legal(
        card_number = 30, 
        max_same_card_number = 31,
        charactor_number = 3,
        check_restriction = False
    )[0]
    assert not deck.check_legal(
        card_number = 30, 
        max_same_card_number = 31,
        charactor_number = 2,
        check_restriction = False
    )[0]
    assert not deck.check_legal(
        card_number = 30, 
        max_same_card_number = 31,
        charactor_number = 4,
        check_restriction = False
    )[0]


def test_deck_restriction():
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Collei
        charactor:Barbara
        Wind and Freedom*30
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert not deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        Wind and Freedom*30
        Sweet Madame*30
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        Elemental Resonance: Woven Waters*30
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        Elemental Resonance: Woven Weeds*30
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert not deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Glorious Season*5
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Grand Expectation*5
    ''')
    assert deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = False,
    )[0]
    assert not deck.check_legal(
        card_number = None, 
        max_same_card_number = None,
        charactor_number = None,
        check_restriction = True,
    )[0]


def test_deck_to_str_to_deck():
    deck = Deck.from_str('''
        default_version:4.0
        charactor:Nahida
        charactor:Mona
        charactor:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Grand Expectation*5
    ''')
    deck2 = Deck.from_str(deck.to_str())
    assert remove_ids(deck) == remove_ids(deck2)


if __name__ == '__main__':
    test_deck_string()
    test_deck_restriction()
    test_deck_to_str_to_deck()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DeclareRoundEndAttackSummonBase

from ...action import Actions, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DieColor, ElementType, 
    FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class SesshouSakura_3_7(DeclareRoundEndAttackSummonBase):
    name: Literal['Sesshou Sakura'] = 'Sesshou Sakura'
    version: Literal['3.7'] = '3.7'
    usage: int = 3
    max_usage: int = 6
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1
    renew_type: Literal['ADD'] = 'ADD'
    extra_attack_usage: int = 4


# Skills


class YakanEvocationSesshouSakura(ElementalSkillBase):
    name: Literal[
        'Yakan Evocation: Sesshou Sakura'] = 'Yakan Evocation: Sesshou Sakura'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon('Sesshou Sakura'),
        ]


class GreatSecretArtTenkoKenshin(ElementalBurstBase):
    name: Literal[
        'Great Secret Art: Tenko Kenshin'] = 'Great Secret Art: Tenko Kenshin'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        summons = match.player_tables[self.position.player_idx].summons
        sakura: SesshouSakura_3_7 | None = None
        for summon in summons:
            if summon.name == 'Sesshou Sakura':
                sakura = summon
                break
        if sakura is not None:
            # destroy and add status
            ret.append(RemoveObjectAction(object_position = sakura.position))
            ret.append(self.create_team_status('Tenko Thunderbolts'))
            if self.is_talent_equipped(match):
                # add status
                ret.append(self.create_charactor_status(
                    "The Shrine's Sacred Shade"))
        return ret


# Talents


class TheShrinesSacredShade_3_7(SkillTalent):
    name: Literal["The Shrine's Sacred Shade"] = "The Shrine's Sacred Shade"
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Yae Miko'] = 'Yae Miko'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: Literal[
        'Great Secret Art: Tenko Kenshin'
    ] = 'Great Secret Art: Tenko Kenshin'


# charactor base


class YaeMiko_3_7(CharactorBase):
    name: Literal['Yae Miko']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | YakanEvocationSesshouSakura
        | GreatSecretArtTenkoKenshin
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Spiritfox Sin-Eater',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            YakanEvocationSesshouSakura(),
            GreatSecretArtTenkoKenshin(),
        ]


register_class(YaeMiko_3_7 | TheShrinesSacredShade_3_7 | SesshouSakura_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase, AttackerSummonBase

from ...event import SkillEndEventArguments

from ...action import (
    Actions, ChangeObjectUsageAction, ChargeAction, CreateObjectAction
)
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DieColor, ElementType, 
    FactionType, ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillBase, SkillTalent
)


# Summons


class Squirrel_3_3(AttackerSummonBase):
    name: Literal['Oceanic Mimic: Squirrel'] = 'Oceanic Mimic: Squirrel'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2


class Raptor_3_3(AttackerSummonBase):
    name: Literal['Oceanic Mimic: Raptor'] = 'Oceanic Mimic: Raptor'
    version: Literal['3.3'] = '3.3'
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1


class Frog_3_3(DefendSummonBase):
    name: Literal['Oceanic Mimic: Frog'] = 'Oceanic Mimic: Frog'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = True


# Skills


mimic_names = [
    'Oceanic Mimic: Squirrel',
    'Oceanic Mimic: Frog',
    'Oceanic Mimic: Raptor'
]


class RhodeiaElementSkill(ElementalSkillBase):
    name: Literal['Oceanid Mimic Summoning',
                  'The Myriad Wilds']
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )
    summon_number: int = 1
    version: Literal['3.3'] = '3.3'

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == 'Oceanid Mimic Summoning':
            pass
        else:
            assert self.name == 'The Myriad Wilds'
            self.summon_number = 2
            self.cost.elemental_dice_number = 5

    def get_next_summon_names(
        self, match: Any, occupied: List[int]
    ) -> int:
        """
        Get next summon names randomly, but fit the rule that try to avoid
        summoning the same type.
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_names = [x.name for x in summons]
        select_idx = []
        for idx, name in enumerate(mimic_names):
            if name not in summon_names and idx not in occupied:
                select_idx.append(idx)
        if len(select_idx) == 0:
            # all exist, get one from not occupied
            select_idx = [x for x in range(3) if x not in occupied]
        if match.config.recreate_mode:
            """
            in recreate mode, the number should always be 1. read name from
            config.

            the key in information is `rhodeia`, name can be one of three
            summons, and can use short names (i.e. frog, raptor, squirrel).
            """
            sname = match.config.random_object_information['rhodeia'].pop(0)
            s_idx = -1
            for idx, name in enumerate(mimic_names):
                if sname.lower() in name.lower():
                    s_idx = idx
                    break
            else:
                raise AssertionError(f'Unknown summon name {sname}')
            assert s_idx in select_idx
            return s_idx
        return select_idx[int(match._random() * len(select_idx))]

    def get_actions(
            self, match: Any) -> List[ChargeAction | CreateObjectAction]:
        """
        create object
        """
        name_idxs = []
        for _ in range(self.summon_number):
            name_idxs.append(self.get_next_summon_names(match, name_idxs))
        ret: List[ChargeAction | CreateObjectAction] = []
        for idx in name_idxs:
            ret.append(self.create_summon(
                mimic_names[idx], { 'version': self.version }
            ))
        ret.append(self.charge_self(1))
        return ret


class TideAndTorrent(ElementalBurstBase):
    name: Literal['Tide and Torrent'] = 'Tide and Torrent'
    damage: int = 2
    damage_per_summon: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        summons = match.player_tables[self.position.player_idx].summons
        self.damage += len(summons) * self.damage_per_summon
        ret = super().get_actions(match)
        self.damage -= len(summons) * self.damage_per_summon
        return ret


# Talents


class StreamingSurge_3_3(SkillTalent):
    name: Literal['Streaming Surge']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Rhodeia of Loch'] = 'Rhodeia of Loch'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4,
        charge = 3
    )
    skill: Literal['Tide and Torrent'] = 'Tide and Torrent'

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments,
        match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        When equipped and this charactor use burst, change all summons' usage
        by 1.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, 
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL,
        ):
            # not equipped, or not self use skill, or not skill
            return []
        skill: SkillBase = match.get_object(event.action.position)
        if skill.skill_type != SkillType.ELEMENTAL_BURST:
            # not elemental burst
            return []
        summons = match.player_tables[self.position.player_idx].summons
        ret = []
        for summon in summons:
            ret.append(ChangeObjectUsageAction(
                object_position = summon.position,
                change_usage = 1
            ))
        return ret


# charactor base


class RhodeiaOfLoch_3_3(CharactorBase):
    name: Literal['Rhodeia of Loch']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase
        | RhodeiaElementSkill | TideAndTorrent
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Surge',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            RhodeiaElementSkill(
                name = 'Oceanid Mimic Summoning',
            ),
            RhodeiaElementSkill(
                name = 'The Myriad Wilds',
            ),
            TideAndTorrent(),
        ]


register_class(
    RhodeiaOfLoch_3_3 | Squirrel_3_3 | Raptor_3_3 | Frog_3_3
    | StreamingSurge_3_3
)



from typing import Any, List, Literal

from ...event import CreateObjectEventArguments, RemoveObjectEventArguments

from ....utils.class_registry import register_class

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class BaneOfAllEvil(ElementalBurstBase):
    name: Literal['Bane of All Evil'] = 'Bane of All Evil'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If has talent, add skill cost decrease by 2
        """
        args = {}
        if self.is_talent_equipped(match):
            args['skill_cost_decrease_usage'] = 2
        return super().get_actions(match) + [
            self.create_charactor_status("Yaksha's Mask", args),
        ]


# Talents


class ConquerorOfEvilGuardianYaksha_3_7(SkillTalent):
    name: Literal['Conqueror of Evil: Guardian Yaksha']
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Xiao'] = 'Xiao'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )
    skill: Literal['Bane of All Evil'] = 'Bane of All Evil'


# charactor base


class Xiao_3_7(CharactorBase):
    name: Literal['Xiao']
    desc: Literal['', 'transform'] = ''
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | BaneOfAllEvil
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Whirlwind Thrust',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Lemniscatic Wind Cycling',
                damage_type = DamageElementalType.ANEMO,
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            BaneOfAllEvil(),
        ]

    def _update_desc(self, match):
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == "Yaksha's Mask":
                self.desc = "transform"
                return
        else:
            self.desc = ''

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []


register_class(Xiao_3_7 | ConquerorOfEvilGuardianYaksha_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AOESummonBase

from ...event import RemoveObjectEventArguments

from ...action import Actions, CreateObjectAction, RemoveObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, CostLabels, DamageElementalType, DieColor, 
    ElementType, FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    CreateStatusPassiveSkill, ElementalBurstBase, ElementalNormalAttackBase, 
    ElementalSkillBase, CharactorBase, TalentBase
)


# Summons


class DarkfireFurnace_3_7(AOESummonBase):
    name: Literal['Darkfire Furnace'] = 'Darkfire Furnace'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = 1
    back_damage: int = 1


# Skills


class OminousStar(ElementalBurstBase):
    name: Literal['Ominous Star'] = 'Ominous Star'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Darkfire Furnace'),
        ]


class FieryRebirth(CreateStatusPassiveSkill):
    name: Literal['Fiery Rebirth'] = 'Fiery Rebirth'
    status_name: Literal['Fiery Rebirth'] = 'Fiery Rebirth'
    regenerate_when_revive: bool = False


# Talents


class EmbersRekindled_3_7(TalentBase):
    name: Literal['Embers Rekindled']
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal[
        'Abyss Lector: Fathomless Flames'] = 'Abyss Lector: Fathomless Flames'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 2
    )
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value 
        | CostLabels.EQUIPMENT.value
    )

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.name == 'Abyss Lector: Fathomless Flames':
                ret.append(c.position)
        return ret

    def is_valid(self, match: Any) -> bool:
        return len(self.get_targets(match)) > 0

    def _attach_status(self, match) -> List[Actions]:
        """
        If self charactor do not have Fiery Rebirth, remove self and attach
        Aegis of Abyssal Flame to charactor.
        """
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        status = charactor.status
        fiery_rebirth_found = False
        for s in status:
            if s.name == 'Fiery Rebirth':
                fiery_rebirth_found = True
                break
        if not fiery_rebirth_found:
            # remove self and attach Aegis of Abyssal Flame to charactor
            return [
                RemoveObjectAction(
                    object_position = self.position,
                ),
                CreateObjectAction(
                    object_position = self.position.set_area(
                        ObjectPositionType.CHARACTOR_STATUS
                    ),
                    object_name = 'Aegis of Abyssal Flame',
                    object_arguments = {}
                ),
            ]
        return []

    def equip(self, match: Any) -> List[Actions]:
        return self._attach_status(match)

    def event_handler_REMOVE_OBJECT(
        self, event: RemoveObjectEventArguments, match: Any
    ) -> List[Actions]:
        if self.position.area != ObjectPositionType.CHARACTOR:
            return []
        return self._attach_status(match)


# charactor base


class AbyssLectorFathomlessFlames_3_7(CharactorBase):
    name: Literal['Abyss Lector: Fathomless Flames']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.PYRO
    max_hp: int = 6
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | ElementalSkillBase | OminousStar 
        | FieryRebirth
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Flame of Salvation',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Searing Precept',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            OminousStar(),
            FieryRebirth(),
        ]


register_class(
    AbyssLectorFathomlessFlames_3_7 | DarkfireFurnace_3_7 | EmbersRekindled_3_7
)



from typing import Any, Dict, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import CostValue
from ...event import (
    ChooseCharactorEventArguments, GameStartEventArguments, 
    RoundPrepareEventArguments, SwitchCharactorEventArguments
)

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    CostLabels, DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, PassiveSkillBase, CharactorBase, TalentBase
)


# Summons


class FrostflakeSekiNoTo_3_3(AttackerSummonBase):
    name: Literal['Frostflake Seki no To'] = 'Frostflake Seki no To'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 2


# Skills


class KamisatoArtSoumetsu(ElementalBurstBase):
    name: Literal['Kamisato Art: Soumetsu'] = 'Kamisato Art: Soumetsu'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Frostflake Seki no To')
        ]


class KamisatoArtSenho(PassiveSkillBase):
    name: Literal['Kamisato Art: Senho'] = 'Kamisato Art: Senho'

    def infuse(self, match: Any) -> List[CreateObjectAction]:
        args: Dict = { 'mark': 'Kamisato Ayaka' }
        if self.is_talent_equipped(match):
            args['talent_activated'] = True
        return [self.create_charactor_status(
            'Cryo Elemental Infusion', args
        )]

    def event_handler_SWITCH_CHARACTOR(
        self, event: SwitchCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.charactor_idx != self.position.charactor_idx
        ):
            # not switch to this charactor
            return []
        return self.infuse(match)

    def event_handler_CHOOSE_CHARACTOR(
        self, event: ChooseCharactorEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.charactor_idx != self.position.charactor_idx
        ):
            # not choose this charactor
            return []
        return self.infuse(match)

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        As after choosing charactor in game start will be followed by a
        round preparing, to make infuse in game start, we increase its usage
        by 1 if it exists.
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        for status in charactor.status:
            if (  # pragma: no branch
                status.name == 'Cryo Elemental Infusion'
                and 'mark' in status.__fields__
                and status.mark == 'Kamisato Ayaka'
            ):
                # add one usage to avoid disappear
                return [ChangeObjectUsageAction(
                    object_position = status.position,
                    change_usage = 1
                )]
        return []


# Talents


class KantenSenmyouBlessing_3_3(TalentBase):
    name: Literal['Kanten Senmyou Blessing']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Kamisato Ayaka'] = 'Kamisato Ayaka'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 2
    )
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.TALENT.value 
        | CostLabels.EQUIPMENT.value
    )
    usage: int = 1
    max_usage: int = 1

    def is_valid(self, match: Any) -> bool:
        if self.position.area != ObjectPositionType.HAND:
            # not in hand, cannot equip
            raise AssertionError('Talent is not in hand')
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        All ayaka are targets
        """
        assert self.position.area != ObjectPositionType.CHARACTOR
        ret: List[ObjectPosition] = []
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.name == self.charactor_name and c.is_alive:
                ret.append(c.position)
        return ret

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        Reset usage
        """
        self.usage = self.max_usage
        return []

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        If equipped, and has usage, and switch to ayaka, reduce 1 cost.
        """
        if self.usage <= 0:
            # out of usage, do nothing
            return value
        if value.cost.label & CostLabels.SWITCH_CHARACTOR.value == 0:
            # not switch charactor, do nothing
            return value
        assert value.target_position is not None
        if not self.position.check_position_valid(
            value.target_position, match, player_idx_same = True,
            charactor_idx_same = True, area_same = True,
            source_area = ObjectPositionType.CHARACTOR,
        ):
            # not equipped or not switch to ayaka, do nothing
            return value
        if value.cost.decrease_cost(None):  # pragma: no branch
            # reduce 1 cost
            if mode == 'REAL':
                self.usage -= 1
        return value


# charactor base


class KamisatoAyaka_3_3(CharactorBase):
    name: Literal['Kamisato Ayaka']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | KamisatoArtSoumetsu
        | KamisatoArtSenho
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Kamisato Art: Kabuki',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Kamisato Art: Hyouka',
                damage_type = DamageElementalType.CRYO,
                cost = ElementalSkillBase.get_cost(self.element)
            ),
            KamisatoArtSoumetsu(),
            KamisatoArtSenho()
        ]


register_class(
    KamisatoAyaka_3_3 | FrostflakeSekiNoTo_3_3 | KantenSenmyouBlessing_3_3
)



from .card import (
    LocationBase, ItemBase, WeaponBase, SupportBase, ArtifactBase, 
    CompanionBase
)
from .charactor import CharactorBase, TalentBase, SkillBase
from .status import StatusBase, CharactorStatusBase, TeamStatusBase
from .summon import SummonBase
from .match import Match, MatchState
from .object_base import ObjectBase, CardBase
from .deck import Deck

# import patch to automatically register classes
from .patch import *  # noqa: F401, F403

__all__ = (
    'LocationBase', 'ItemBase', 'WeaponBase', 'SupportBase', 'ArtifactBase', 
    'CompanionBase', 'CharactorBase', 'TalentBase', 'SkillBase', 
    'StatusBase', 'CharactorStatusBase', 'TeamStatusBase', 'SummonBase',
    'Match', 'MatchState', 'ObjectBase', 'CardBase', 'Deck'
)



from .server.deck import Deck  # noqa: F401
from .server.match import Match, MatchState  # noqa: F401


try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ModuleNotFoundError:
    __version__ = 'unknown'
    __version_tuple__ = (0, 0, 0)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import SwirlChangeSummonBase

from ...event import RoundEndEventArguments

from ...action import (
    ActionTypes, Actions, ChangeObjectUsageAction, MakeDamageAction
)
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class Stormeye_3_7(SwirlChangeSummonBase):
    name: Literal['Stormeye'] = 'Stormeye'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage: int = 2

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        Need to change enemy active
        """
        ret = super().event_handler_ROUND_END(event, match)
        our_active = match.player_tables[
            self.position.player_idx].active_charactor_idx
        target_charactors = match.player_tables[
            1 - self.position.player_idx].charactors
        target_idx = our_active
        if target_charactors[target_idx].is_defeated:
            # need to choose another charactor
            for idx, charactor in enumerate(target_charactors):
                if not charactor.is_defeated:
                    target_idx = idx
                    break
            else:
                raise AssertionError('No charactor alive')
        attack_action = ret[0]
        assert attack_action.type == ActionTypes.MAKE_DAMAGE
        attack_action.charactor_change_idx[
            1 - self.position.player_idx] = target_idx
        return ret


# Skills


class SkywardSonnet(ElementalSkillBase):
    name: Literal['Skyward Sonnet'] = 'Skyward Sonnet'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        assert charactor.name == 'Venti'
        talent_activated = False
        if self.is_talent_equipped(match):
            talent_activated = True
        return super().get_actions(match) + [
            self.create_team_status(
                'Stormzone', 
                {'talent_activated': talent_activated}
            )
        ]


class WindsGrandOde(ElementalBurstBase):
    name: Literal["Wind's Grand Ode"] = "Wind's Grand Ode"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        # create summon first, so it can change element immediately.
        return [self.create_summon('Stormeye')] + super().get_actions(match)


# Talents


class EmbraceOfWinds_3_7(SkillTalent):
    name: Literal['Embrace of Winds']
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Venti'] = 'Venti'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )
    skill: Literal['Skyward Sonnet'] = 'Skyward Sonnet'


# charactor base


class Venti_3_7(CharactorBase):
    name: Literal['Venti']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SkywardSonnet | WindsGrandOde
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Divine Marksmanship',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SkywardSonnet(),
            WindsGrandOde()
        ]


register_class(Venti_3_7 | EmbraceOfWinds_3_7 | Stormeye_3_7)



from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....action import CreateDiceAction

from ....event import SkillEndEventArguments
from ....consts import (
    ELEMENT_TO_DIE_COLOR, ElementType, ObjectPositionType, SkillType, 
    WeaponType
)
from ....struct import Cost
from .base import RoundEffectWeaponBase


class SacrificialWeapons_3_3(RoundEffectWeaponBase):
    name: Literal[
        'Sacrificial Fragments',
        'Sacrificial Greatsword',
        'Sacrificial Sword',
        'Sacrificial Bow',
    ]
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.3'] = '3.3'
    weapon_type: WeaponType = WeaponType.OTHER
    max_usage_per_round: int = 1

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == 'Sacrificial Fragments':
            self.weapon_type = WeaponType.CATALYST
        elif self.name == 'Sacrificial Bow':
            self.weapon_type = WeaponType.BOW
        elif self.name == 'Sacrificial Sword':
            self.weapon_type = WeaponType.SWORD
        else:
            assert self.name == 'Sacrificial Greatsword'
            self.weapon_type = WeaponType.CLAYMORE

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        If self use elemental skill, create one corresponding dice
        """
        if not (
            self.position.area == ObjectPositionType.CHARACTOR
            and event.action.position.player_idx == self.position.player_idx 
            and event.action.position.charactor_idx 
            == self.position.charactor_idx
            and event.action.skill_type == SkillType.ELEMENTAL_SKILL
            and self.usage > 0
        ):
            # not equipped or not self use elemental skill or no usage
            return []
        self.usage -= 1
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        ele_type: ElementType = charactor.element
        die_color = ELEMENT_TO_DIE_COLOR[ele_type]
        return [CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 1,
            color = die_color,
        )]


register_class(SacrificialWeapons_3_3)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ..charactor_base import ElementalSkillBase, PhysicalNormalAttackBase

from ...action import Actions

from ...consts import DamageElementalType, DieColor

from ...struct import Cost

from .tartaglia_4_1 import (
    HavocObliteration as HO_4_1,
    TideWithholder as TW_4_1,
    Tartaglia_4_1 as Tartaglia_4_1,
    AbyssalMayhemHydrospout_4_1 as AMH_4_1,
)


# Skills


class FoulLegacyRagingTide(ElementalSkillBase):
    name: Literal['Foul Legacy: Raging Tide'] = 'Foul Legacy: Raging Tide'
    version: Literal['3.7'] = '3.7' 
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object then attack
        """
        return [
            self.create_charactor_status(
                'Melee Stance',
                { 'version': self.version }
            ),
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]


class HavocObliteration(HO_4_1):
    name: Literal['Havoc: Obliteration'] = 'Havoc: Obliteration'
    version: Literal['3.7'] = '3.7' 
    ranged_damage: int = 4


class TideWithholder(TW_4_1):
    version: Literal['3.7'] = '3.7'


# Talents


class AbyssalMayhemHydrospout_3_7(AMH_4_1):
    version: Literal['3.7']
    only_active_charactor: bool = False
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4,
    )


# charactor base


class Tartaglia_3_7(Tartaglia_4_1):
    name: Literal['Tartaglia']
    version: Literal['3.7']
    skills: List[
        PhysicalNormalAttackBase | FoulLegacyRagingTide
        | HavocObliteration | TideWithholder
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Cutting Torrent',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            FoulLegacyRagingTide(),
            HavocObliteration(),
            TideWithholder(),
        ]


register_class(Tartaglia_3_7 | AbyssalMayhemHydrospout_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, SkillType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class GardenOfPurity_3_6(AttackerSummonBase):
    name: Literal['Garden of Purity'] = 'Garden of Purity'
    version: Literal['3.6'] = '3.6'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            assert mode == 'REAL'
            value.damage += 1
        return value


# Skills


class KamisatoArtKyouka(ElementalSkillBase):
    name: Literal['Kamisato Art: Kyouka'] = 'Kamisato Art: Kyouka'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_charactor_status(
                'Takimeguri Kanka',
                {
                    'usage': 3,
                    'max_usage': 3
                }
            ),
        ]


class KamisatoArtSuiyuu(ElementalBurstBase):
    name: Literal['Kamisato Art: Suiyuu'] = 'Kamisato Art: Suiyuu'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon('Garden of Purity')
        ]


# Talents


class KyoukaFuushi_3_6(SkillTalent):
    name: Literal['Kyouka Fuushi']
    version: Literal['3.6'] = '3.6'
    charactor_name: Literal['Kamisato Ayato'] = 'Kamisato Ayato'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )
    skill: Literal['Kamisato Art: Kyouka'] = 'Kamisato Art: Kyouka'


# charactor base


class KamisatoAyato_4_1(CharactorBase):
    name: Literal['Kamisato Ayato']
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | KamisatoArtKyouka | KamisatoArtSuiyuu
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Kamisato Art: Marobashi',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            KamisatoArtKyouka(),
            KamisatoArtSuiyuu(),
        ]


register_class(KamisatoAyato_4_1 | KyoukaFuushi_3_6 | GardenOfPurity_3_6)



"""
Collect descriptions.
"""
import json
import logging
from typing import Any, Dict, Literal, TypedDict


ExpectedLanguageType = Literal['zh-CN', 'en-US']


class DescDictType(TypedDict, total = False):
    """
    image_path: optional, image path
    id: optional, id for sorting cards
    names: name of the card, key is language type, value is name
    descs: description of the card, key is version, value is dict of language
        type and description
    cost: cost of the card, key is version, value is dict of Cost type
    """
    image_path: str
    id: int
    names: Dict[ExpectedLanguageType, str]
    descs: Dict[str, Dict[ExpectedLanguageType, str]]
    cost: Dict[str, Dict[str, Any]]


_default_json_path = __file__[:-len('desc_registry.py')] + 'default_desc.json'
_desc_dict: Dict[str, DescDictType] = json.load(
    open(_default_json_path, 'r', encoding = 'utf-8')
)


def _parse_value(value: Any):
    """
    If value is str and starts with $: it is a reference to current desc, 
    return its referenced result. Reference key format is:
    $key1|key2|key3 ...
    """
    if isinstance(value, str) and value.startswith('$'):
        # return referenced result
        keys = value[1:].split('|')
        root = _desc_dict
        for key in keys:
            if key not in root:
                raise ValueError(f'in "{value}": key "{key}" not found')
            root = root[key]
        return root
    return value


def _merge_dict(source, destination, prev_key: str = ''):
    for key, value in source.items():
        if prev_key == '':
            pkey = key
        else:
            pkey = f'{prev_key}|{key}'
        if isinstance(value, dict):
            # get node or create one
            node = destination.setdefault(key, {})
            _merge_dict(value, node, prev_key = pkey)
        else:
            value = _parse_value(value)
            if '|names|' in pkey and value in ['', 'XXX']:
                # name is empty, raise error.
                raise ValueError(
                    f'name value "{value}" of key "{pkey}" is empty'
                )
            if key in destination:
                if destination[key] != value:
                    raise ValueError(
                        f'value "{destination[key]}" of key "{pkey}" '
                        f'is replaced by new value "{value}"'
                    )
                else:
                    logging.warning(f'duplicate adding key "{pkey}"')
            destination[key] = value
    return destination


def update_desc(desc_dict: Dict[str, DescDictType]) -> None:
    """
    Update descriptions.
    """
    _merge_dict(desc_dict, _desc_dict)


def desc_exist(type: str, name: str, version: str) -> bool:
    """
    Check if the description of the class exists.
    """
    full_name = f'{type}/{name}'
    return (
        full_name in _desc_dict 
        and 'descs' in _desc_dict[full_name]
        and version in _desc_dict[full_name]['descs']  # type: ignore
    )


def get_desc_patch() -> Dict[str, DescDictType]:
    """
    Get desc patch.
    """
    return _desc_dict


def update_cost(type: str, name: str, version: str, cost: Any) -> None:
    """
    update cost into desc dict.
    """
    if not desc_exist(type, name, version):
        raise ValueError(f'cannot find desc for {type}/{name}/{version}')
    full_name = f'{type}/{name}'
    d = _desc_dict[full_name]
    if 'cost' not in _desc_dict[full_name]:
        d['cost'] = {}
    d['cost'][version] = cost.dict()  # type: ignore


__all__ = [
    'DescDictType', 'ExpectedLanguageType', 'update_desc', 'desc_exist',
    'get_desc_patch', 'update_cost'
]



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear
# Skills


class ChonghuasLayeredFrost(ElementalSkillBase):
    name: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {}
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        talent = charactor.talent
        if talent is not None:
            args = {
                'talent_activated': True,
                'usage': talent.status_max_usage,
                'max_usage': talent.status_max_usage
            }
        return super().get_actions(match) + [
            self.create_team_status("Chonghua's Frost Field", args),
        ]


# Talents


class SteadyBreathing_3_3(SkillTalent):
    name: Literal['Steady Breathing']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Chongyun'] = 'Chongyun'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 4
    )
    skill: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    status_max_usage: int = 3


class SteadyBreathing_4_2(SkillTalent):
    name: Literal['Steady Breathing']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Chongyun'] = 'Chongyun'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 3
    )
    skill: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    status_max_usage: int = 2


# charactor base


class Chongyun_3_3(CharactorBase):
    name: Literal['Chongyun']  # Do not set default value for charactor name
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | ChonghuasLayeredFrost | ElementalBurstBase
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Demonbane',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ChonghuasLayeredFrost(),
            ElementalBurstBase(
                name = 'Cloud-Parting Star',
                damage = 7,
                damage_type = DamageElementalType.CRYO,
                cost = ElementalBurstBase.get_cost(self.element, 3, 3)
            )
        ]


register_class(Chongyun_3_3 | SteadyBreathing_3_3 | SteadyBreathing_4_2)





from typing import Any, List, Literal

from .....utils.class_registry import register_class

from ....modifiable_values import DamageIncreaseValue

from ....action import CreateObjectAction

from ....event import SkillEndEventArguments

from .base import RoundEffectWeaponBase, WeaponBase

from ....struct import Cost, ObjectPosition

from ....consts import ObjectPositionType, ObjectType, WeaponType


class WolfsGravestone_3_3(WeaponBase):
    name: Literal["Wolf's Gravestone"] = "Wolf's Gravestone"
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.3'] = '3.3'
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        assert value.target_position.area == ObjectPositionType.CHARACTOR
        charactor = match.get_object(value.target_position)
        if charactor.hp <= 6:
            self.damage_increase = 3
        super().value_modifier_DAMAGE_INCREASE(value, match, mode)
        self.damage_increase = 1
        return value


class TheBell_3_7(RoundEffectWeaponBase):
    name: Literal['The Bell']
    type: Literal[ObjectType.WEAPON] = ObjectType.WEAPON
    version: Literal['3.7'] = '3.7'
    weapon_type: WeaponType = WeaponType.CLAYMORE

    cost: Cost = Cost(same_dice_number = 3)
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self charactor use any skill, and have usage, create Rebellious
        Shield.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL
        ):
            # not self charactor use skill
            return []
        if self.usage == 0:
            # no usage
            return []
        self.usage -= 1
        return [CreateObjectAction(
            object_name = 'Rebellious Shield',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = -1
            ),
            object_arguments = {}
        )]


register_class(WolfsGravestone_3_3 | TheBell_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    CharactorBase, PhysicalNormalAttackBase, SkillTalent
)


class Breastplate(ElementalSkillBase):
    name: Literal['Breastplate'] = 'Breastplate'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.GEO
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_team_status('Full Plate'))
        return ret


class SweepingTime(ElementalBurstBase):
    name: Literal['Sweeping Time'] = 'Sweeping Time'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_number = 4,
        elemental_dice_color = DieColor.GEO,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        ret.append(self.create_charactor_status('Sweeping Time'))
        return ret


class IGotYourBack_3_3(SkillTalent):
    name: Literal['I Got Your Back']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Noelle'] = 'Noelle'
    cost: Cost = Cost(
        elemental_dice_number = 3,
        elemental_dice_color = DieColor.GEO
    )
    skill: Literal['Breastplate'] = 'Breastplate'


class Noelle_3_3(CharactorBase):
    name: Literal['Noelle']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Breastplate | SweepingTime
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Favonius Bladework - Maid',
                cost = PhysicalNormalAttackBase.get_cost(
                    ElementType.GEO,
                )
            ),
            Breastplate(),
            SweepingTime()
        ]


register_class(Noelle_3_3 | IGotYourBack_3_3)
